import asyncio
import datetime
import json
import sys
from argparse import ArgumentParser
from pathlib import Path

from stac_generator.__version__ import __version__
from stac_generator.base.schema import StacCatalogConfig, StacCollectionConfig
from stac_generator.generator_factory import StacGeneratorFactory


def run_cli() -> None:
    # Build the CLI argument parser
    parser = ArgumentParser(prog="stac_generator", description="CLI tool to generator STAC records")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    # Source commands
    parser.add_argument("data_type", type=str, help="data type of the source data")
    parser.add_argument("source_file", type=str, help="path to the source config csv")
    parser.add_argument(
        "--to_remote",
        type=str,
        required=False,
        default=None,
        help="catalog api endpoint for pushing generated stac records",
    )
    parser.add_argument(
        "--to_local",
        type=str,
        required=False,
        default=None,
        help="local path to save generated stac records to",
    )

    # Collection Information
    collection_metadata = parser.add_argument_group("STAC collection metadata")
    collection_metadata.add_argument("--id", type=str, help="id of collection")
    collection_metadata.add_argument("--title", type=str, help="title of collection", required=False, default="Auto-generated.")
    collection_metadata.add_argument(
        "--description",
        type=str,
        help="description of collection",
        required=False,
        default="Auto-generated",
    )

    # Catalog Information
    catalog_metadata = parser.add_argument_group("STAC catalog metadata- Derived from STAC Collection metadata if not provided")
    catalog_metadata.add_argument(
        "--catalog_id",
        type=str,
        required=False,
        help="id of catalog. Use the value of id if not provided",
    )
    catalog_metadata.add_argument(
        "--catalog_title",
        type=str,
        required=False,
        help="title of catalog. Use the value of title if not provided",
    )
    catalog_metadata.add_argument(
        "--catalog_description",
        type=str,
        required=False,
        help="description of catalog. Use the value of description if not provided",
    )

    # STAC Common Metadata
    common_metadata = parser.add_argument_group(
        "STAC common metadata. Providers information can only be provided from a metadata json file. CLI arguments take priority over json fields if overlapping occurs"
    )
    common_metadata.add_argument(
        "--datetime",
        type=datetime.datetime.fromisoformat,
        help="STAC datetime",
        required=False,
        default=None,
    )
    common_metadata.add_argument(
        "--start_datetime",
        type=datetime.datetime.fromisoformat,
        help="STAC start_datetime",
        required=False,
        default=None,
    )
    common_metadata.add_argument(
        "--end_datetime",
        type=datetime.datetime.fromisoformat,
        help="STAC end_datetime",
        required=False,
        default=None,
    )
    common_metadata.add_argument("--license", type=str, help="STAC license", required=False, default="proprietary")
    common_metadata.add_argument("--platform", type=str, help="STAC platform", required=False, default=None)
    common_metadata.add_argument("--constellation", type=str, help="STAC constellation", required=False, default=None)
    common_metadata.add_argument("--mission", type=str, help="STAC mission", required=False, default=None)
    common_metadata.add_argument("--gsd", type=float, help="STAC gsd", required=False, default=None)
    common_metadata.add_argument(
        "--instruments",
        action="extend",
        nargs="+",
        type=str,
        required=False,
        default=None,
        help="STAC instrument",
    )
    common_metadata.add_argument(
        "--metadata_json",
        type=str,
        required=False,
        default=None,
        help="path to json file describing the metadata",
    )
    args = parser.parse_args()

    if args.to_remote is None and args.to_local is None:
        sys.exit("Error - either --to_remote or --to_local must be provided")

    # Build collection config and catalog config
    metadata_json = {}
    if args.metadata_json:
        with Path(args.metadata_json).open("r") as file:
            metadata_json = json.load(file)

    # CLI args take precedence over metadata fields
    collection_config = StacCollectionConfig(
        id=args.id,
        title=args.title,
        description=args.description,
        datetime=args.datetime if args.datetime else metadata_json.get("datetime"),
        start_datetime=args.start_datetime if args.start_datetime else metadata_json.get("start_datetime"),
        end_datetime=args.end_datetime if args.end_datetime else metadata_json.get("end_datetime"),
        license=args.license if args.license else metadata_json.get("license"),
        platform=args.platform if args.platform else metadata_json.get("platform"),
        constellation=args.constellation if args.constellation else metadata_json.get("constellation"),
        mission=args.mission if args.mission else metadata_json.get("mission"),
        gsd=args.gsd if args.gsd else metadata_json.get("gsd"),
        instruments=args.instruments if args.instruments else metadata_json.get("instruments"),
        providers=metadata_json.get("providers"),
    )
    # CLI catalog args take precendence over collection args
    catalog_config = StacCatalogConfig(
        id=args.catalog_id if args.catalog_id else args.id,
        title=args.catalog_title if args.catalog_title else args.title,
        description=args.catalog_description if args.catalog_description else args.description,
    )
    # Generate
    generator = StacGeneratorFactory.get_stac_generator(
        data_type=args.data_type,
        source_file=args.source_file,
        collection_cfg=collection_config,
        catalog_cfg=catalog_config,
    )
    # Save
    if args.to_remote:
        asyncio.run(generator.write_to_api(href=args.to_remote))
        print(f"Catalog successfully pushed to {args.to_remote}")
    if args.to_local:
        generator.generate_catalog_and_save(href=args.to_local)
        print(f"Catalog successfully save to {args.to_local}")


if __name__ == "__main__":
    run_cli()



import pandas as pd
import pystac

from stac_generator._types import CsvMediaType
from stac_generator.base.generator import StacGenerator
from stac_generator.base.schema import StacCatalogConfig, StacCollectionConfig
from stac_generator.csv.driver import CsvDriver
from stac_generator.csv.schema import CsvConfig, CsvExtension
from stac_generator.csv.utils import group_df, items_from_group_df


class CsvGenerator(StacGenerator[CsvConfig]):
    def __init__(
        self,
        source_df: pd.DataFrame,
        collection_cfg: StacCollectionConfig,
        catalog_cfg: StacCatalogConfig | None = None,
        href: str | None = None,
    ) -> None:
        super().__init__(
            source_df=source_df,
            collection_cfg=collection_cfg,
            catalog_cfg=catalog_cfg,
            href=href,
            driver=CsvDriver,
        )
        self.driver: type[CsvDriver]

    def create_item_from_config(self, source_cfg: CsvConfig) -> list[pystac.Item]:
        asset = pystac.Asset(
            href=source_cfg.location,
            description="Raw csv data",
            roles=["data"],
            media_type=CsvMediaType,
        )
        raw_df = self.driver(source_cfg).get_data()
        group_map = group_df(raw_df, source_cfg.prefix, source_cfg.groupby)
        properties = CsvExtension.model_validate(source_cfg, from_attributes=True).model_dump()
        return items_from_group_df(
            group_map,
            asset,
            source_cfg.epsg,
            source_cfg.T,
            source_cfg.datetime,
            source_cfg.start_datetime,
            source_cfg.end_datetime,
            properties,
        )



import datetime as pydatetime
import json
from collections.abc import Sequence
from itertools import chain
from typing import Any, cast

import geopandas as gpd
import pandas as pd
import pystac
from pystac.extensions.projection import ItemProjectionExtension
from shapely import MultiPoint, Point, to_geojson

from stac_generator._types import TimeExtentT
from stac_generator.csv.schema import ColumnInfo


def read_csv(
    src_path: str,
    X_coord: str,
    Y_coord: str,
    T_coord: str | None = None,
    date_format: str = "ISO8601",
    columns: list[str] | list[ColumnInfo] | None = None,
    groupby: list[str] | None = None,
) -> pd.DataFrame:
    """Read in csv from local disk

    Users must provide at the bare minimum the location of the csv, and the names of the columns to be
    treated as the X and Y coordinates. By default, will read in all columns in the csv. If columns and groupby
    columns are provided, will selectively read specified columns together with the coordinate columns (X, Y, T).

    :param src_path: path to csv file
    :type src_path: str
    :param X_coord: name of X field
    :type X_coord: str
    :param Y_coord: name of Y field
    :type Y_coord: str
    :param T_coord: name of time field, defaults to None
    :type T_coord: str | None, optional
    :param date_format: format to pass to pandas to parse datetime, defaults to "ISO8601"
    :type date_format: str, optional
    :param columns: band information, defaults to None
    :type columns: list[str] | list[ColumnInfo] | None, optional
    :param groupby: list of fields that partition the points into groups, defaults to None
    :type groupby: list[str] | None, optional
    :return: read dataframe
    :rtype: pd.DataFrame
    """
    parse_dates: list[str] | bool = [T_coord] if isinstance(T_coord, str) else False
    usecols: list[str] | None = None
    # If band info is provided, only read in the required columns + the X and Y coordinates
    if columns:
        if isinstance(columns[0], str):
            usecols = cast(list[str], list(columns))
        else:
            usecols = [item["name"] for item in cast(list[ColumnInfo], columns)]
        usecols.extend([X_coord, Y_coord])
        if T_coord:
            usecols.append(T_coord)
        # If item group provided -> read in
        if groupby:
            usecols.extend(groupby)
    return pd.read_csv(
        filepath_or_buffer=src_path,
        usecols=usecols,
        date_format=date_format,
        parse_dates=parse_dates,
    )


def to_gdf(df: pd.DataFrame, X_coord: str, Y_coord: str, epsg: int) -> gpd.GeoDataFrame:
    return gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df[X_coord], df[Y_coord], crs=epsg))


def calculate_temporal_extent(
    df: gpd.GeoDataFrame | None = None,
    time_col: str | None = None,
    datetime: pydatetime.datetime | None = None,
    start_datetime: pydatetime.datetime | None = None,
    end_datetime: pydatetime.datetime | None = None,
) -> TimeExtentT:
    """Get temporal extent based on Stac specification.

    Use `start_datetime` and `end_datetime` if provided. Otherwise, use `datetime` if provided.
    If the dataframe and time column are provided, the function will obtain the start datetime and end datetime from the dataframe.

    :param df: Provided dataframe.
    :type df: gpd.GeoDataFrame | None, optional

    :param time_col: Name of time column.
    :type time_col: str | None, optional

    :param datetime: Datetime.
    :type datetime: datetime.datetime | None, optional

    :param start_datetime: Start datetime.
    :type start_datetime: datetime.datetime | None, optional

    :param end_datetime: End datetime.
    :type end_datetime: datetime.datetime | None, optional

    :return: Start datetime, end datetime.
    :rtype: TimeExtentT
    """
    if start_datetime and end_datetime:
        return start_datetime, end_datetime
    if datetime:
        return datetime, datetime
    if df is not None and isinstance(time_col, str):
        if time_col not in df.columns:
            raise KeyError(f"Cannot find time_col: {time_col} in given dataframe")
        if not isinstance(df[time_col].dtype, pydatetime.datetime):
            raise ValueError(
                f"Dtype of time_col: {time_col} must be of datetime type: {df[time_col].dtype}"
            )
        min_T, max_T = df[time_col].min(), df[time_col].max()
        return (min_T, max_T)
    raise ValueError(
        "If datetime is None, both start_datetime and end_datetime values must be provided"
    )


def calculate_geometry(
    df: gpd.GeoDataFrame,
) -> Point | MultiPoint:
    """Calculate the geometry from geopandas dataframe.

    Work only on point based data

    Returns a `shapely.Point` or `shapely.MultiPoint` depending on the number
    of unique points in the dataframe.

    :param df: source dataframe
    :type df: gpd.GeoDataFrame
    :return: shapely geometry object
    :rtype: Point | MultiPoint
    """
    points: Sequence[Point] = df["geometry"].unique()
    if len(points) == 1:
        return points[0]
    return MultiPoint([[p.x, p.y] for p in points])


def group_df(
    df: gpd.GeoDataFrame,
    prefix: str,
    groupby: Sequence[str] | None = None,
) -> dict[str, gpd.GeoDataFrame]:
    """Partition dataframe into sub-dataframes based on fields in `groupby`.
    Each partition will be assigned an item name obtained from the collection name and field values.

    For example, if the collection name is `point_data` and `groupby = ["sites"]` with values being `["A", "B"]`,
    the resulting item names will be `point_data_site_A_item` and `point_data_site_B_item`.

    If `groupby` is not provided, return a single collection item which is the full dataframe.

    :param df: Source dataframe.
    :type df: gpd.GeoDataFrame

    :param prefix: Prefix for each item ID.
    :type prefix: str

    :param groupby: Fields to partition the dataframe. Must be present in the original dataframe.
    :type groupby: Sequence[str] | None, optional

    :return: Mapping of group name to sub-dataframe.
    :rtype: dict[str, gpd.GeoDataFrame]
    """
    if not groupby:
        item_name = prefix
        return {item_name: df}
    partition_df = df.groupby(groupby).apply(lambda group: group.drop_duplicates())  # type: ignore[call-overload]
    partition_df = partition_df.reset_index(level=-1, drop=True)
    df_group = {}
    for i in range(len(partition_df)):
        idx = (
            partition_df.index[i] if len(groupby) != 1 else [partition_df.index[i]]
        )  # If groupby has one single item, convert idx to list of 1
        group_name = "_".join([str(item) for item in chain(*zip(groupby, idx, strict=True))])
        item_name = f"{prefix}_{group_name}"
        df_group[item_name] = partition_df.loc[idx, :].reset_index(drop=True)
    return df_group


def items_from_group_df(
    group_df: dict[str, gpd.GeoDataFrame],
    asset: pystac.Asset,
    epsg: int,
    T: str | None = None,
    datetime: pydatetime.datetime | None = None,
    start_datetime: pydatetime.datetime | None = None,
    end_datetime: pydatetime.datetime | None = None,
    properties: dict[str, Any] | None = None,
) -> list[pystac.Item]:
    """Extract `shapely.Point` data from partitioned dataframe maps.

    This function takes a group_df - i.e. dictionary of point group name and their
    dataframe (as obtained from `stac_generator.csv.utils.group_df` method) to generate
    a list of pystac.Item based on information from the dataframe.

    :param group_df: dictionary of point group and their source csv
    :type group_df: dict[str, gpd.GeoDataFrame]
    :param asset: source data asset
    :type asset: pystac.Asset
    :param epsg: epsg code of the source dataframe
    :type epsg: int
    :param T: name of the time column, defaults to None
    :type T: str | None, optional
    :param datetime: pystac datetime metadata, defaults to None
    :type datetime: datetime.datetime | None, optional
    :param start_datetime: pystac start_datetime metadata, defaults to None
    :type start_datetime: datetime.datetime | None, optional
    :param end_datetime: pystac end_datetime metadata, defaults to None
    :type end_datetime: datetime.datetime | None, optional
    :param properties: additional properties to be added to item
    :type properties: dict[str, Any] | None, optional
    :return: list of generated stac items
    :rtype: list[pystac.Item]
    """
    _properties = properties if properties else {}
    assets = {"source": asset}
    items = []
    for item_id, item_df in group_df.items():
        _start_datetime, _end_datetime = calculate_temporal_extent(
            item_df, T, datetime, start_datetime, end_datetime
        )
        _start_datetime = _start_datetime if _start_datetime is not None else start_datetime
        _end_datetime = _end_datetime if _end_datetime is not None else end_datetime
        _datetime = datetime if datetime else _end_datetime
        _geometry = json.loads(to_geojson(calculate_geometry(item_df)))
        item = pystac.Item(
            item_id,
            bbox=item_df.total_bounds.tolist(),
            geometry=_geometry,
            datetime=_datetime,
            start_datetime=_start_datetime,
            end_datetime=_end_datetime,
            properties=_properties,
            assets=assets,
        )
        proj_ext = ItemProjectionExtension.ext(item, add_if_missing=True)
        proj_ext.apply(epsg=epsg)
        items.append(item)
    return items



import geopandas as gpd
import pandas as pd

from stac_generator.base.driver import IODriver
from stac_generator.csv.schema import CsvConfig
from stac_generator.csv.utils import read_csv, to_gdf


class CsvDriver(IODriver):
    def __init__(self, config: CsvConfig) -> None:
        super().__init__(config)
        self.config: CsvConfig

    def get_data(self) -> gpd.GeoDataFrame:
        return self.read_local()

    def _to_gdf(self, df: pd.DataFrame) -> gpd.GeoDataFrame:
        return to_gdf(df, self.config.X, self.config.Y, self.config.epsg)

    def read_local(self) -> gpd.GeoDataFrame:
        assert self.config.local is not None
        df = read_csv(
            self.config.local,
            self.config.X,
            self.config.Y,
            self.config.T,
            self.config.date_format,
            self.config.column_info,
            self.config.groupby,
        )
        return self._to_gdf(df)



import datetime
from typing import Any, cast

import geopandas as gpd
import pystac
from shapely import (
    Geometry,
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)

from stac_generator.base.schema import StacCollectionConfig


def geometry_from_dict(item: dict[str, Any]) -> Geometry:
    """Create a `shapely.Geometry` object from dictionary

    :param item: dictionary that must conform to geojson <a href="https://datatracker.ietf.org/doc/html/rfc7946">geometry object</a>
    :type item: dict[str, Any]
    :raises ValueError: if item is None
    :raises ValueError: if no `type` key is found in item
    :raises ValueError: if `type` is not `GeometryCollection` but no `coordinates` key is found in item
    :raises ValueError: if `type` is `GeometryCollection` but no `geometries` key is found in item
    :raises ValueError: if `type` is not a valid geojson geometry type from rfc7946
    :return: geometry item as `shapely.Geometry` object
    :rtype: Geometry
    """
    if not item:
        raise ValueError("Expects non null geometry")
    if "type" not in item:
        raise ValueError("Invalid geojson geometry object: no type field")
    if item["type"] != "GeometryCollection" and "coordinates" not in item:
        raise ValueError(
            "Invalid geojson geometry object: no coordinates field for non GeometryCollection"
        )
    geometry_type = item["type"]
    coordinates = item.get("coordinates")
    match geometry_type:
        case "Point":
            return Point(cast(list, coordinates))
        case "MultiPoint":
            return MultiPoint(cast(list, coordinates))
        case "LineString":
            return LineString(*cast(list, coordinates))
        case "MultiLineString":
            return MultiLineString(cast(list, coordinates))
        case "Polygon":
            return Polygon(cast(list, coordinates))
        case "MultiPolygon":
            return MultiPolygon(cast(list, coordinates))
        case "GeometryCollection":
            if "geometries" not in item:
                raise ValueError(
                    "Invalid geojson geometry object: no geometries field for GeometryCollection type"
                )
            return GeometryCollection(cast(list, item["geometries"]))
        case _:
            raise ValueError(f"Invalid geojson geometry type: {geometry_type}")


def extract_spatial_extent(items: list[pystac.Item]) -> pystac.SpatialExtent:
    """Extract spatial extent for a collection from child items

    :param items: list of all `pystac.Item`
    :type items: list[pystac.Item]
    :return: spatial extent object
    :rtype: pystac.SpatialExtent
    """
    geometries: list[Geometry] = []
    for item in items:
        if (geo := item.geometry) is not None:
            geometries.append(geometry_from_dict(geo))
    geo_series = gpd.GeoSeries(data=geometries)
    bbox = geo_series.total_bounds.tolist()
    return pystac.SpatialExtent(bbox)


def extract_temporal_extent(
    items: list[pystac.Item], collection: StacCollectionConfig | None = None
) -> pystac.TemporalExtent:
    """Extract spatial extent for a collection from a list of items and collection config.

    If temporal extent (`start_datetime`, `end_datetime` or `datetime`) is in `collection`, generate
    `pystac.TemporalExtent` from those fields. Otherwise, extract the fields from the provided items.

    :param items: list of Items
    :type items: list[pystac.Item]
    :param collection: collection config, defaults to None
    :type collection: StacCollectionConfig | None, optional
    :raises ValueError: if a pystac.Item has neither `datetime` nor both `start_datetime` and `end_datetime`
    :return: extracted temporal extent
    :rtype: pystac.TemporalExtent
    """
    if collection:
        if collection.start_datetime and collection.end_datetime:
            return pystac.TemporalExtent([[collection.start_datetime, collection.end_datetime]])
        return pystac.TemporalExtent([collection.datetime, collection.datetime])
    min_dt = datetime.datetime.now(datetime.UTC)
    max_dt = datetime.datetime(1, 1, 1)  # noqa: DTZ001
    for item in items:
        if "start_datetime" in item.properties and "end_datetime" in item.properties:
            min_dt = min(min_dt, item.properties["start_datetime"])
            max_dt = max(max_dt, item.properties["end_datetime"])
        else:
            if item.datetime is None:
                raise ValueError(
                    "Invalid pystac item. Either datetime or start_datetime and end_datetime values must be provided"
                )
            min_dt = min(min_dt, item.datetime)
            max_dt = max(max_dt, item.datetime)
    max_dt = max(max_dt, min_dt)
    return pystac.TemporalExtent([[min_dt, max_dt]])



from __future__ import annotations

import datetime
from typing import Any, Self, TypeVar, cast

from httpx._types import (
    RequestData,  # noqa: TCH002
)
from pydantic import BaseModel, model_validator
from stac_pydantic.shared import StacCommonMetadata as _StacCommonMetaData

from stac_generator._types import (  # noqa: TCH001
    CookieTypes,
    HeaderTypes,
    HTTPMethod,
    QueryParamTypes,
    RequestContent,
    StacEntityT,
)

T = TypeVar("T", bound="SourceConfig")


class StacCommonMetadata(_StacCommonMetaData):
    """Stac Common Metadata. Automatically sets datetime values to
    current datetime if neither datetime nor start_datetime and end_datetime are provided
    """

    @model_validator(mode="before")
    @classmethod
    def set_datetime(self, data: Any) -> Any:
        if isinstance(data, dict) and (
            ("datetime" not in data or data["datetime"] is None)
            and ("start_datetime" not in data or data["start_datetime"] is None)
            and ("end_datetime" not in data or data["end_datetime"] is None)
        ):
            now = datetime.datetime.now(datetime.UTC)
            data["datetime"] = now
            data["start_datetime"] = now
            data["end_datetime"] = now
        return data


class StacCatalogConfig(BaseModel):
    """Contains parameters to pass to Catalog constructor"""

    id: str
    """Catalog id"""
    title: str
    """Catalog title"""
    description: str = "Auto-generated Stac Catalog"
    """Catalog description"""


class StacCollectionConfig(StacCommonMetadata):
    """Contains parameters to pass to Collection constructor. Also contains other metadata

    This config provides additional information that can not be derived from source file, which includes
    <a href="https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md">Stac Common Metadata</a>
    and other descriptive information such as the id of the new entity
    """

    # Stac Information
    id: str
    """Collection id"""
    title: str
    """Collection title"""
    description: str = "Auto-generated Stac Collection"
    """Collection description"""


class StacItemConfig(StacCommonMetadata):
    """Contains parameters to pass to Item constructor.

    This config provides additional information that can not be derived from source file, which includes
    <a href="https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md">Stac Common Metadata</a>
    and other descriptive information such as the id of the new entity
    """

    prefix: str
    """Item prefix - doubles as ID if there is only one item extracted from the source file"""
    description: str = "Auto-generated Stac Item"
    """Item description"""


class SourceConfig(StacItemConfig):
    """Base source config that should be subclassed for different file extensions.

    Source files contain raw spatial information (i.e. geotiff, shp, csv) from which
    some Stac metadata can be derived. SourceConfig describes:

    - The access mechanisms for the source file: stored on local disk, or hosted somewhere behind an api endpoint. If the source
    file must be accessed through an endpoint, users can provide additional HTTP information that forms the HTTP request to the host server.
    - Processing information that are unique for the source type. Users should inherit `SourceConfig` for file extensions
    currently unsupported.
    - Additional Stac Metadata from `StacConfig`
    """

    location: str
    """Asset's href. If `local` is not provided, will be used as endpoint for fetching data file, using other config parameters.
    If the file is on disk, it is acceptable to set location as ./path/to/source/file
    """
    local: str | None = None
    """Path to the source file on local disk. This is for reading in and processing local file.
    If local is not provided, HTTP methods and other parameters must be provided for reading in the file from `location` field"""
    extension: str | None = None
    """Explicit file extension specification. If the file is stored behind an api endpoint, the field `extension` must be provided"""
    # HTTP Parameters
    method: HTTPMethod | None = "GET"
    """HTTPMethod to acquire the file from `location`"""
    params: QueryParamTypes | None = None
    """HTTP query params for getting file from `location`"""
    headers: HeaderTypes | None = None
    """HTTP query headers for getting file from `location`"""
    cookies: CookieTypes | None = None
    """HTTP query cookies for getting file from `location`"""
    content: RequestContent | None = None
    """HTTP query body content for getting file from `location`"""
    data: RequestData | None = None
    """HTTP query body content for getting file from `location`"""
    json_body: Any = None
    """HTTP query body content for getting file from `location`"""

    @property
    def source_extension(self) -> str:
        if self.extension:
            return self.extension
        return (cast(str, self.local)).split(".")[-1]

    @model_validator(mode="after")
    def validate_require_extension_when_source_is_remote(self) -> Self:
        if self.local is None and self.extension is None:
            raise ValueError(
                "If source is must be accessed through an endpoint, extension field must be specified"
            )
        return self


class LoadConfig(BaseModel):
    entity: StacEntityT
    """Stac Entity type - Item, ItemCollection, Collection, Catalog"""
    json_location: str | None = None
    """Stac Json file on disk"""
    stac_api_endpoint: str | None = None
    """Stac API Endpoint"""

    @model_validator(mode="after")
    def validate_field(self) -> Self:
        if self.json_location is None and self.stac_api_endpoint is None:
            raise ValueError("One of json_location or stac_api_endpoint field must be not None")
        return self



import abc
from typing import Any

import httpx

from stac_generator.base.schema import SourceConfig


class IODriver:
    """Base driver for handling IO for different file formats/extensions

    The primary purpose of this class is to enable reading source file from location specified in
    the source config. File can either be read locally or via sending a request to the hosting server.
    """

    def __init__(self, config: SourceConfig) -> None:
        self.config = config

    @abc.abstractmethod
    def get_data(self) -> Any:
        """Read data based on config information"""
        raise NotImplementedError

    @staticmethod
    def fetch(config: SourceConfig) -> httpx.Response:
        """Use config information to send a GET request to an endpoint

        :param config: config information containing the url and request headers
        :type config: SourceConfig
        :return: raw Response object
        :rtype: httpx.Response
        """
        return httpx.request(
            method=str(config.method),
            url=str(config.location),
            params=config.params,
            headers=config.headers,
            json=config.json_body,
            cookies=config.cookies,
            content=config.content,
            data=config.data,
        )



import abc
from typing import Any, Generic, cast

import httpx
import pandas as pd
import pystac
from pystac.collection import Extent

from stac_generator._types import StacEntityT
from stac_generator.base.driver import IODriver
from stac_generator.base.schema import (
    LoadConfig,
    StacCatalogConfig,
    StacCollectionConfig,
    T,
)
from stac_generator.base.utils import extract_spatial_extent, extract_temporal_extent


class StacGenerator(Generic[T]):
    source_type: type[T]
    """SourceConfig subclass that contains information used for parsing the source file"""

    @classmethod
    def __class_getitem__(cls, source_type: type) -> type:
        kwargs = {"source_type": source_type}
        return type(f"StacGenerator[{source_type.__name__}]", (StacGenerator,), kwargs)

    def __init__(
        self,
        source_df: pd.DataFrame,
        collection_cfg: StacCollectionConfig,
        catalog_cfg: StacCatalogConfig | None = None,
        href: str | None = None,
        driver: type[IODriver] | None = None,
    ) -> None:
        """Base STACGenerator object. Users should extend this class for handling different file extensions.
        Please see CsvStacGenerator source code.

        :param source_df: source data config
        :type source_df: pd.DataFrame
        :param collection_cfg: collection config for generating collection. Required if the intention is to save the catalog to disk or POST to an API, defaults to None
        :type collection_cfg: StacCollectionConfig
        :param catalog_cfg: catalog config for generating catalog. Required if the intention is to save the catalog to disk or POST to an API, defaults to None
        :type catalog_cfg: StacCatalogConfig | None, optional
        :param href: catalog href for serialisation.
        :type href: str | None, optional
        :param driver: driver that handles T file extension using config for extension type T. Note that this field is not required
        :type driver: type[IODriver] | None, optional

        """
        self.collection_cfg = collection_cfg
        self.catalog_cfg = (
            catalog_cfg
            if catalog_cfg
            else StacCatalogConfig(
                id=collection_cfg.id,
                title=collection_cfg.title,
                description=collection_cfg.description,
            )
        )
        self.source_df = source_df
        self.driver = driver
        self.configs = [
            self.source_type(**cast(dict[str, Any], self.source_df.loc[i, :].to_dict()))
            for i in range(len(self.source_df))
        ]
        self.href = href

    @abc.abstractmethod
    def create_item_from_config(self, source_cfg: T) -> list[pystac.Item]:
        """Abstract method that handles `pystac.Item` generation from the appropriate config"""
        raise NotImplementedError

    @staticmethod
    def extract_extent(
        items: list[pystac.Item], collection_cfg: StacCollectionConfig | None = None
    ) -> Extent:
        return Extent(extract_spatial_extent(items), extract_temporal_extent(items, collection_cfg))

    def _create_collection_from_items(
        self,
        items: list[pystac.Item],
        collection_cfg: StacCollectionConfig | None = None,
    ) -> pystac.Collection:
        if collection_cfg is None:
            raise ValueError("Generating collection requires non null collection config")
        collection = pystac.Collection(
            id=collection_cfg.id,
            description=(
                collection_cfg.description
                if collection_cfg.description
                else f"Auto-generated collection {collection_cfg.id} with stac_generator"
            ),
            extent=self.extract_extent(items, collection_cfg),
            title=collection_cfg.title,
            license=collection_cfg.license if collection_cfg.license else "proprietary",
            providers=[
                pystac.Provider.from_dict(item.model_dump()) for item in collection_cfg.providers
            ]
            if collection_cfg.providers
            else None,
        )
        collection.add_items(items)
        return collection

    def _create_catalog_from_collection(
        self, collection: pystac.Collection, catalog_cfg: StacCatalogConfig | None = None
    ) -> pystac.Catalog:
        if catalog_cfg is None:
            raise ValueError("Generating catalog requires non null catalog config")
        catalog = pystac.Catalog(
            id=catalog_cfg.id,
            description=catalog_cfg.description,
            title=catalog_cfg.title,
        )
        catalog.add_child(collection)
        return catalog

    def create_items(self) -> list[pystac.Item]:
        """Generate STAC Items from source dataframe

        :return: list of generated STAC Item
        :rtype: list[pystac.Item]
        """
        items = []
        for config in self.configs:
            items.extend(self.create_item_from_config(config))
        return items

    def create_collection(self) -> pystac.Collection:
        """Generate STAC Collection that includes all STAC Items from source dataframe

        collection_cfg must be provided to constructor for this method to work

        :return: STAC Collection that includes all generated STAC Items
        :rtype: pystac.Collection
        """
        items = self.create_items()
        return self._create_collection_from_items(items, self.collection_cfg)

    def create_catalog(self) -> pystac.Catalog:
        """Generate STAC Catalog that includes a STAC Collection containing all STAC Item generated from
        source dataframe

        Both collection_cfg and catalog_cfg must be provided for this method to work

        :return: STAC Catalog
        :rtype: pystac.Catalog
        """
        collection = self.create_collection()
        return self._create_catalog_from_collection(collection, self.catalog_cfg)

    def generate_collection_and_save(self, href: str) -> None:
        """Generate STAC Collection and save to disk.

        This is a convenient method
        that combines `self.create_collection` and `collection.normalize_and_save(href)`


        :param href: disk location of the generated collection
        :type href: str
        """
        collection = self.create_collection()
        collection.normalize_hrefs(href)
        collection.validate_all()
        collection.save()

    def generate_catalog_and_save(self, href: str | None) -> None:
        """Write the catalog generated from source dataframe to local disk

        If `href` is not provided, will attemp to use `href` under `StacCatalogConfig`

        :param href: disk location
        :type href: str | None
        :raises ValueError: if href parameter is not provided as method argument or config attribute
        """
        if not href and not (self.catalog_cfg and self.href):
            raise ValueError("Href must be provided as argument or under catalog config")
        catalog = self.create_catalog()
        href = href if href is not None else cast(str, self.href)
        catalog.normalize_hrefs(href)
        catalog.validate_all()
        catalog.save()

    async def write_to_api(self, href: str | None = None) -> None:
        """Write the catalog generated from source dataframe to an href

        If href is not provided, will attempt to use the href under `StacCatalogConfig`

        :param href: STAC api href, defaults to None
        :type href: str | None, optional
        :raises ValueError: if href parameter is not provided as method argument or config attribute
        """
        if not href and not (self.catalog_cfg and self.href):
            raise ValueError("href must be provided as argument or to catalog config")
        catalog = self.create_catalog()
        href = href if href is not None else cast(str, self.href)
        catalog.normalize_hrefs(href)
        catalog.validate_all()
        async with httpx.AsyncClient() as client:
            for collection in catalog.get_collections():
                for item in collection.get_all_items():
                    await client.post(
                        f"{href}/collection/{item.collection_id}/items/{item.id}",
                        json=item.to_dict(),
                    )
                await client.post(f"{href}/collection/{collection.id}", json=collection.to_dict())


class StacLoader:
    @staticmethod
    def _load_from_file(
        entity: StacEntityT, location: str
    ) -> pystac.Catalog | pystac.Collection | pystac.Item | pystac.ItemCollection:
        match entity:
            case "Item":
                return pystac.Item.from_file(location)
            case "ItemCollection":
                return pystac.ItemCollection.from_file(location)
            case "Collection":
                return pystac.Collection.from_file(location)
            case "Catalogue":
                return pystac.Catalog.from_file(location)
            case _:
                raise ValueError(f"Invalid Stac type: {entity}")

    @staticmethod
    def _load_from_api(
        entity: StacEntityT, api_href: str
    ) -> pystac.Catalog | pystac.Collection | pystac.Item | pystac.ItemCollection:
        response = httpx.get(api_href)
        json_data = response.json()
        match entity:
            case "Item":
                return pystac.Item(**json_data)
            case "ItemCollection":
                return pystac.ItemCollection(**json_data)
            case "Collection":
                return pystac.Collection(**json_data)
            case "Catalogue":
                return pystac.Catalog(**json_data)
            case _:
                raise ValueError(f"Invalid Stac type: {entity}")

    @staticmethod
    def from_config(
        load_cfg: LoadConfig,
    ) -> pystac.Catalog | pystac.Collection | pystac.Item | pystac.ItemCollection:
        """Get a Stac entity using information from load_cfg

        :param load_cfg: load information
        :type load_cfg: LoadConfig
        :return: pystac entity
        :rtype: pystac.Catalog | pystac.Collection | pystac.Item | pystac.ItemCollection
        """
        if load_cfg.json_location:
            return StacLoader._load_from_file(load_cfg.entity, load_cfg.json_location)
        return StacLoader._load_from_api(load_cfg.entity, cast(str, load_cfg.stac_api_endpoint))
